use axum::http::StatusCode;
use axum::response::{IntoResponse, Response};
use axum::Json;
use magnetar_calckey_model::{CalckeyCacheError, CalckeyDbError};
use serde::Serialize;
use serde_json::json;

pub mod auth;

#[derive(Debug, Clone, Serialize)]
#[repr(transparent)]
pub struct ErrorCode(pub String);

// This janky hack allows us to use `.error_code()` on enums with strum::IntoStaticStr
pub trait IntoErrorCode {
    fn error_code<'a, 'b: 'a>(&'a self) -> ErrorCode
    where
        &'a Self: Into<&'b str>;
}

impl<T: ?Sized> IntoErrorCode for T {
    fn error_code<'a, 'b: 'a>(&'a self) -> ErrorCode
    where
        &'a Self: Into<&'b str>,
    {
        ErrorCode(self.into().to_string())
    }
}

impl ErrorCode {
    pub fn join(&self, other: &str) -> Self {
        Self(format!("{}:{}", other, self.0))
    }
}

#[derive(Debug)]
pub struct ApiError {
    pub status: StatusCode,
    pub code: ErrorCode,
    pub message: String,
}

impl IntoResponse for ApiError {
    fn into_response(self) -> Response {
        (
            self.status,
            Json(json!({
                "status": self.status.as_u16(),
                "code": self.code,
                "message": self.message,
            })),
        )
            .into_response()
    }
}

impl From<CalckeyDbError> for ApiError {
    fn from(err: CalckeyDbError) -> Self {
        Self {
            status: StatusCode::INTERNAL_SERVER_ERROR,
            code: err.error_code(),
            message: if cfg!(debug_assertions) {
                format!("Database error: {}", err)
            } else {
                "Database error".to_string()
            },
        }
    }
}

impl From<CalckeyCacheError> for ApiError {
    fn from(err: CalckeyCacheError) -> Self {
        Self {
            status: StatusCode::INTERNAL_SERVER_ERROR,
            code: err.error_code(),
            message: if cfg!(debug_assertions) {
                format!("Cache error: {}", err)
            } else {
                "Cache error".to_string()
            },
        }
    }
}
